@@ -33,11 +33,12 @@
 # limitations under the License.
 """Kerberos security provider"""
 import logging
+import shlex
 import socket
 import subprocess
 import sys
 import time
-from typing import Optional
+from typing import List, Optional
 
 from airflow.configuration import conf
 
@@ -46,7 +47,7 @@ NEED_KRB181_WORKAROUND = None  # type: Optional[bool]
 log = logging.getLogger(__name__)
 
 
-def renew_from_kt(principal: str, keytab: str, exit_on_fail: bool = True):
+def renew_from_kt(principal: Optional[str], keytab: str, exit_on_fail: bool = True):
     """
     Renew kerberos token from keytab
 
@@ -58,20 +59,34 @@ def renew_from_kt(principal: str, keytab: str, exit_on_fail: bool = True):
     # minutes to give ourselves a large renewal buffer.
     renewal_lifetime = f"{conf.getint('kerberos', 'reinit_frequency')}m"
 
-    cmd_principal = principal or conf.get('kerberos', 'principal').replace("_HOST", socket.getfqdn())
+    cmd_principal = principal or conf.get_mandatory_value('kerberos', 'principal').replace(
+        "_HOST", socket.getfqdn()
+    )
 
-    cmdv = [
-        conf.get('kerberos', 'kinit_path'),
+    if conf.getboolean('kerberos', 'forwardable'):
+        forwardable = '-f'
+    else:
+        forwardable = '-F'
+
+    if conf.getboolean('kerberos', 'include_ip'):
+        include_ip = '-a'
+    else:
+        include_ip = '-A'
+
+    cmdv: List[str] = [
+        conf.get_mandatory_value('kerberos', 'kinit_path'),
+        forwardable,
+        include_ip,
         "-r",
         renewal_lifetime,
         "-k",  # host ticket
         "-t",
         keytab,  # specify keytab
         "-c",
-        conf.get('kerberos', 'ccache'),  # specify credentials cache
+        conf.get_mandatory_value('kerberos', 'ccache'),  # specify credentials cache
         cmd_principal,
     ]
-    log.info("Re-initialising kerberos from keytab: %s", " ".join(cmdv))
+    log.info("Re-initialising kerberos from keytab: %s", " ".join(shlex.quote(f) for f in cmdv))
 
     with subprocess.Popen(
         cmdv,
@@ -94,14 +109,14 @@ def renew_from_kt(principal: str, keytab: str, exit_on_fail: bool = True):
             else:
                 return subp.returncode
 
-    global NEED_KRB181_WORKAROUND  # pylint: disable=global-statement
+    global NEED_KRB181_WORKAROUND
     if NEED_KRB181_WORKAROUND is None:
         NEED_KRB181_WORKAROUND = detect_conf_var()
     if NEED_KRB181_WORKAROUND:
         # (From: HUE-640). Kerberos clock have seconds level granularity. Make sure we
         # renew the ticket after the initial valid time.
         time.sleep(1.5)
-        ret = perform_krb181_workaround(principal)
+        ret = perform_krb181_workaround(cmd_principal)
         if exit_on_fail and ret != 0:
             sys.exit(ret)
         else:
@@ -116,10 +131,10 @@ def perform_krb181_workaround(principal: str):
     :param principal: principal name
     :return: None
     """
-    cmdv = [
-        conf.get('kerberos', 'kinit_path'),
+    cmdv: List[str] = [
+        conf.get_mandatory_value('kerberos', 'kinit_path'),
         "-c",
-        conf.get('kerberos', 'ccache'),
+        conf.get_mandatory_value('kerberos', 'ccache'),
         "-R",
     ]  # Renew ticket_cache
 
@@ -129,17 +144,16 @@ def perform_krb181_workaround(principal: str):
 
     if ret != 0:
         principal = f"{principal or conf.get('kerberos', 'principal')}/{socket.getfqdn()}"
-        princ = principal
-        ccache = conf.get('kerberos', 'principal')
+        ccache = conf.get('kerberos', 'ccache')
         log.error(
             "Couldn't renew kerberos ticket in order to work around Kerberos 1.8.1 issue. Please check that "
             "the ticket for '%s' is still renewable:\n  $ kinit -f -c %s\nIf the 'renew until' date is the "
             "same as the 'valid starting' date, the ticket cannot be renewed. Please check your KDC "
             "configuration, and the ticket renewal policy (maxrenewlife) for the '%s' and `krbtgt' "
             "principals.",
-            princ,
+            principal,
             ccache,
-            princ,
+            principal,
         )
     return ret
 
@@ -150,14 +164,14 @@ def detect_conf_var() -> bool:
     Sun Java Krb5LoginModule in Java6, so we need to take an action to work
     around it.
     """
-    ticket_cache = conf.get('kerberos', 'ccache')
+    ticket_cache = conf.get_mandatory_value('kerberos', 'ccache')
 
     with open(ticket_cache, 'rb') as file:
         # Note: this file is binary, so we check against a bytearray.
         return b'X-CACHECONF:' in file.read()
 
 
-def run(principal: str, keytab: str):
+def run(principal: Optional[str], keytab: str):
     """
     Run the kerbros renewer.
 
@@ -166,7 +180,7 @@ def run(principal: str, keytab: str):
     :return: None
     """
     if not keytab:
-        log.debug("Keytab renewer not starting, no keytab configured")
+        log.warning("Keytab renewer not starting, no keytab configured")
         sys.exit(0)
 
     while True:
