@@ -17,12 +17,21 @@
 # under the License.
 
 from collections import Counter
+from typing import TYPE_CHECKING
 
+from sqlalchemy import func
+
+from airflow.ti_deps.dep_context import DepContext
 from airflow.ti_deps.deps.base_ti_dep import BaseTIDep
-from airflow.utils.session import provide_session
+from airflow.utils.session import NEW_SESSION, provide_session
 from airflow.utils.state import State
 from airflow.utils.trigger_rule import TriggerRule as TR
 
+if TYPE_CHECKING:
+    from sqlalchemy.orm import Session
+
+    from airflow.models.taskinstance import TaskInstance
+
 
 class TriggerRuleDep(BaseTIDep):
     """
@@ -35,17 +44,15 @@ class TriggerRuleDep(BaseTIDep):
     IS_TASK_DEP = True
 
     @staticmethod
-    def _get_states_count_upstream_ti(ti, finished_tasks):
+    def _get_states_count_upstream_ti(task, finished_tis):
         """
         This function returns the states of the upstream tis for a specific ti in order to determine
         whether this ti can run in this iteration
 
         :param ti: the ti that we want to calculate deps for
-        :type ti: airflow.models.TaskInstance
-        :param finished_tasks: all the finished tasks of the dag_run
-        :type finished_tasks: list[airflow.models.TaskInstance]
+        :param finished_tis: all the finished tasks of the dag_run
         """
-        counter = Counter(task.state for task in finished_tasks if task.task_id in ti.task.upstream_task_ids)
+        counter = Counter(ti.state for ti in finished_tis if ti.task_id in task.upstream_task_ids)
         return (
             counter.get(State.SUCCESS, 0),
             counter.get(State.SKIPPED, 0),
@@ -55,7 +62,7 @@ class TriggerRuleDep(BaseTIDep):
         )
 
     @provide_session
-    def _get_dep_statuses(self, ti, session, dep_context):
+    def _get_dep_statuses(self, ti, session, dep_context: DepContext):
         # Checking that all upstream dependencies have succeeded
         if not ti.task.upstream_list:
             yield self._passing_status(reason="The task instance did not have any upstream tasks.")
@@ -66,7 +73,7 @@ class TriggerRuleDep(BaseTIDep):
             return
         # see if the task name is in the task upstream for our task
         successes, skipped, failed, upstream_failed, done = self._get_states_count_upstream_ti(
-            ti=ti, finished_tasks=dep_context.ensure_finished_tasks(ti.get_dagrun(session), session)
+            task=ti.task, finished_tis=dep_context.ensure_finished_tis(ti.get_dagrun(session), session)
         )
 
         yield from self._evaluate_trigger_rule(
@@ -80,37 +87,62 @@ class TriggerRuleDep(BaseTIDep):
             session=session,
         )
 
+    @staticmethod
+    def _count_upstreams(ti: "TaskInstance", *, session: "Session"):
+        from airflow.models.taskinstance import TaskInstance
+
+        # Optimization: Don't need to hit the database if no upstreams are mapped.
+        upstream_task_ids = ti.task.upstream_task_ids
+        if ti.task.dag and not any(ti.task.dag.get_task(tid).is_mapped for tid in upstream_task_ids):
+            return len(upstream_task_ids)
+
+        # We don't naively count task instances because it is not guaranteed
+        # that all upstreams have been created in the database at this point.
+        # Instead, we look for already-expanded tasks, and add them to the raw
+        # task count without considering mapping.
+        mapped_tis_addition = (
+            session.query(func.count())
+            .filter(
+                TaskInstance.dag_id == ti.dag_id,
+                TaskInstance.run_id == ti.run_id,
+                TaskInstance.task_id.in_(upstream_task_ids),
+                TaskInstance.map_index > 0,
+            )
+            .scalar()
+        )
+        return len(upstream_task_ids) + mapped_tis_addition
+
     @provide_session
     def _evaluate_trigger_rule(
-        self, ti, successes, skipped, failed, upstream_failed, done, flag_upstream_failed, session
+        self,
+        ti,
+        successes,
+        skipped,
+        failed,
+        upstream_failed,
+        done,
+        flag_upstream_failed,
+        session: "Session" = NEW_SESSION,
     ):
         """
         Yields a dependency status that indicate whether the given task instance's trigger
         rule was met.
 
         :param ti: the task instance to evaluate the trigger rule of
-        :type ti: airflow.models.TaskInstance
         :param successes: Number of successful upstream tasks
-        :type successes: int
         :param skipped: Number of skipped upstream tasks
-        :type skipped: int
         :param failed: Number of failed upstream tasks
-        :type failed: int
         :param upstream_failed: Number of upstream_failed upstream tasks
-        :type upstream_failed: int
         :param done: Number of completed upstream tasks
-        :type done: int
         :param flag_upstream_failed: This is a hack to generate
             the upstream_failed state creation while checking to see
             whether the task instance is runnable. It was the shortest
             path to add the feature
-        :type flag_upstream_failed: bool
         :param session: database session
-        :type session: sqlalchemy.orm.session.Session
         """
         task = ti.task
-        upstream = len(task.upstream_task_ids)
-        trigger_rule = task.trigger_rule
+        upstream = self._count_upstreams(ti, session=session)
+        trigger_rule: TR = task.trigger_rule
         upstream_done = done >= upstream
         upstream_tasks_state = {
             "total": upstream,
@@ -154,82 +186,101 @@ class TriggerRuleDep(BaseTIDep):
             elif trigger_rule == TR.NONE_SKIPPED:
                 if skipped:
                     ti.set_state(State.SKIPPED, session)
+            elif trigger_rule == TR.ALL_SKIPPED:
+                if successes or failed:
+                    ti.set_state(State.SKIPPED, session)
 
         if trigger_rule == TR.ONE_SUCCESS:
             if successes <= 0:
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires one upstream "
-                    "task success, but none were found. "
-                    "upstream_tasks_state={}, upstream_task_ids={}".format(
-                        trigger_rule, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires one upstream task success, "
+                        f"but none were found. upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         elif trigger_rule == TR.ONE_FAILED:
             if not failed and not upstream_failed:
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires one upstream "
-                    "task failure, but none were found. "
-                    "upstream_tasks_state={}, upstream_task_ids={}".format(
-                        trigger_rule, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires one upstream task failure, "
+                        f"but none were found. upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         elif trigger_rule == TR.ALL_SUCCESS:
             num_failures = upstream - successes
             if num_failures > 0:
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires all upstream "
-                    "tasks to have succeeded, but found {} non-success(es). "
-                    "upstream_tasks_state={}, upstream_task_ids={}".format(
-                        trigger_rule, num_failures, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
+                        f"succeeded, but found {num_failures} non-success(es). "
+                        f"upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         elif trigger_rule == TR.ALL_FAILED:
             num_successes = upstream - failed - upstream_failed
             if num_successes > 0:
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires all upstream "
-                    "tasks to have failed, but found {} non-failure(s). "
-                    "upstream_tasks_state={}, upstream_task_ids={}".format(
-                        trigger_rule, num_successes, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have failed, "
+                        f"but found {num_successes} non-failure(s). "
+                        f"upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         elif trigger_rule == TR.ALL_DONE:
             if not upstream_done:
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires all upstream "
-                    "tasks to have completed, but found {} task(s) that "
-                    "were not done. upstream_tasks_state={}, "
-                    "upstream_task_ids={}".format(
-                        trigger_rule, upstream_done, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
+                        f"completed, but found {upstream_done} task(s) that were not done. "
+                        f"upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         elif trigger_rule == TR.NONE_FAILED:
             num_failures = upstream - successes - skipped
             if num_failures > 0:
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires all upstream "
-                    "tasks to have succeeded or been skipped, but found {} non-success(es). "
-                    "upstream_tasks_state={}, upstream_task_ids={}".format(
-                        trigger_rule, num_failures, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
+                        f"succeeded or been skipped, but found {num_failures} non-success(es). "
+                        f"upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         elif trigger_rule == TR.NONE_FAILED_MIN_ONE_SUCCESS:
             num_failures = upstream - successes - skipped
             if num_failures > 0:
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires all upstream "
-                    "tasks to have succeeded or been skipped, but found {} non-success(es). "
-                    "upstream_tasks_state={}, upstream_task_ids={}".format(
-                        trigger_rule, num_failures, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
+                        f"succeeded or been skipped, but found {num_failures} non-success(es). "
+                        f"upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         elif trigger_rule == TR.NONE_SKIPPED:
             if not upstream_done or (skipped > 0):
                 yield self._failing_status(
-                    reason="Task's trigger rule '{}' requires all upstream "
-                    "tasks to not have been skipped, but found {} task(s) skipped. "
-                    "upstream_tasks_state={}, upstream_task_ids={}".format(
-                        trigger_rule, skipped, upstream_tasks_state, task.upstream_task_ids
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to not have been "
+                        f"skipped, but found {skipped} task(s) skipped. "
+                        f"upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
+                    )
+                )
+        elif trigger_rule == TR.ALL_SKIPPED:
+            num_non_skipped = upstream - skipped
+            if num_non_skipped > 0:
+                yield self._failing_status(
+                    reason=(
+                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have been "
+                        f"skipped, but found {num_non_skipped} task(s) in non skipped state. "
+                        f"upstream_tasks_state={upstream_tasks_state}, "
+                        f"upstream_task_ids={task.upstream_task_ids}"
                     )
                 )
         else:
