@@ -18,14 +18,15 @@
 import collections
 import logging
 import re
-from typing import TYPE_CHECKING, Iterable, Optional, Set, TypeVar, Union
+from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Tuple, Union
 
+from airflow import settings
 from airflow.compat.functools import cache, cached_property
 
 if TYPE_CHECKING:
     from airflow.typing_compat import RePatternType
 
-    RedactableItem = TypeVar('RedactableItem')
+    RedactableItem = Union[str, Dict[Any, Any], Tuple[Any, ...], List[Any]]
 
 
 log = logging.getLogger(__name__)
@@ -43,10 +44,14 @@ DEFAULT_SENSITIVE_FIELDS = frozenset(
         'private_key',
         'secret',
         'token',
+        'keyfile_dict',
+        'service_account',
     }
 )
 """Names of fields (Connection extra, Variable key name etc.) that are deemed sensitive"""
 
+SECRETS_TO_SKIP_MASKING_FOR_TESTS = {'airflow'}
+
 
 @cache
 def get_sensitive_variables_fields():
@@ -70,7 +75,7 @@ def should_hide_value_for_key(name):
     return False
 
 
-def mask_secret(secret: Union[str, dict, Iterable], name: str = None) -> None:
+def mask_secret(secret: Union[str, dict, Iterable], name: Optional[str] = None) -> None:
     """
     Mask a secret from appearing in the task logs.
 
@@ -80,18 +85,15 @@ def mask_secret(secret: Union[str, dict, Iterable], name: str = None) -> None:
     If ``secret`` is a dict or a iterable (excluding str) then it will be
     recursively walked and keys with sensitive names will be hidden.
     """
-    # Delay import
-    from airflow import settings
-
     # Filtering all log messages is not a free process, so we only do it when
     # running tasks
-    if not settings.MASK_SECRETS_IN_LOGS or not secret:
+    if not secret:
         return
 
     _secrets_masker().add_mask(secret, name)
 
 
-def redact(value: "RedactableItem", name: str = None) -> "RedactableItem":
+def redact(value: "RedactableItem", name: Optional[str] = None) -> "RedactableItem":
     """Redact any secrets found in ``value``."""
     return _secrets_masker().redact(value, name)
 
@@ -145,13 +147,21 @@ class SecretsMasker(logging.Filter):
         return frozenset(record.__dict__).difference({'msg', 'args'})
 
     def _redact_exception_with_context(self, exception):
-        exception.args = (self.redact(v) for v in exception.args)
+        # Exception class may not be modifiable (e.g. declared by an
+        # extension module such as JDBC).
+        try:
+            exception.args = (self.redact(v) for v in exception.args)
+        except AttributeError:
+            pass
         if exception.__context__:
             self._redact_exception_with_context(exception.__context__)
         if exception.__cause__ and exception.__cause__ is not exception.__context__:
             self._redact_exception_with_context(exception.__cause__)
 
     def filter(self, record) -> bool:
+        if settings.MASK_SECRETS_IN_LOGS is not True:
+            return True
+
         if self.ALREADY_FILTERED_FLAG in record.__dict__:
             # Filters are attached to multiple handlers and logs, keep a
             # "private" flag that stops us needing to process it more than once
@@ -232,13 +242,16 @@ class SecretsMasker(logging.Filter):
         """
         return self._redact(item, name, depth=0)
 
-    def add_mask(self, secret: Union[str, dict, Iterable], name: str = None):
+    def add_mask(self, secret: Union[str, dict, Iterable], name: Optional[str] = None):
         """Add a new secret to be masked to this filter instance."""
+        from airflow.configuration import conf
+
+        test_mode: bool = conf.getboolean('core', 'unit_test_mode')
         if isinstance(secret, dict):
             for k, v in secret.items():
                 self.add_mask(v, k)
         elif isinstance(secret, str):
-            if not secret:
+            if not secret or (test_mode and secret in SECRETS_TO_SKIP_MASKING_FOR_TESTS):
                 return
             pattern = re.escape(secret)
             if pattern not in self.patterns and (not name or should_hide_value_for_key(name)):
